#coding=utf-8

###################################################################
# Copyright 2013 Civa lin 林雪凡
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#      http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
###################################################################

"""
LoNote intermediation 中介格式模組

定義 LoNote 標準資料格式 "Book"，並提供操作的 API。

=================================

IM 的大尺度操作：

    1. 新建一個 Book 物件，同時指定 url 新建或讀取
    3. 對資料進行各種存讀處理 ...
    4. 用 .close() 結束操作

---------------------------------

IM 系統的設計邏輯：

    1. 整體可類比為一本書 (Book)。
    2. 書中有許多紙張 (Paper)。
        - 紙張本身是單獨存在的
        - 紙張與其他紙張的關係，如排序與嵌套方式，會被放在 Book 層級中。
    3. 紙張中含有多個部份 (sections) 可供獨立書寫。
        - 如正面、背面、對話、補充、備份等。
    4. 以紙張為單位，能夠加入附加資源檔案 (resources)，受個別紙張管理。

=================================

IM 細節：

    1. 所有文字資料採用 utf-8 編碼
    2. 所有時間資料字串採用以下格式：
        - TimeStandard.format == '%Y-%m-%dT%H:%M:%S.%f%z'
    3. 所有時間資料都含有時區訊息

=================================

函式命名規則如下：

    1. 名詞在前動詞在後
        - 如：content_update(...)
    2. 只有名詞表示讀取
        - 如：content(...)
    3. 只有動詞表示針對整體行動
        - 如：update(...)

    常用綴尾動詞包括：
        create = 新建一筆資料。和已有資料衝突會拋出 IMAreadyExist
        update = 更新一筆資料，如果資料不存在會自動新建。
                 （但有少數函式無法自動新建，此時會拋出例外）
                 操作失敗會拋出 IMError 底下的各種錯誤
        delete = 表示刪除一筆資料，資料不存在時會拋出 IMNotExist
        flush  = 將最新版本的資料寫入底層。
                 「底層」是何義由個案決定。不過通常是指一個較慢的操作。
        assert = 斷言某事成立，否則會拋出錯誤

    各種開頭：
        is     = 進行判斷，回傳 True or False

參數命名法表記如下：

    url      = (str) 實體資料路徑，像是網址或本地檔案位址等。
    tar      = (lonote.tar3.Tar) 已開啟的 Tar 物件，作為底層資料存取途徑。

    id       > 供機器使用的 id，在各自的命名空間內需為唯一，不可衝突。
               對資料進行各種操作，需透過這些 id 來定位要操作哪一個。

               *注意!* ：id 需要為「合法檔名」！否則可能出錯

               包括以下幾種……（只在容易混淆時細分）

        bid  = (str) 書籍 id
        pid  = (str) 紙張 id
        sid  = (str) section id
        rid  = (str) 資源 id
        xid  = (str) 任意 id

    ctime    = (datetime) 建立時間
    mtime    = (datetime) 修改時間

    name     > 給人類閱讀用的名字，命名無限制，也可以任意重名。

        bname = (str) 書籍標題（同封面紙張標題）
        pname = (str) 紙張標題

    desc     > 給人類閱讀的說明或是副標。

        bdesc = (str) 書籍說明（同封面紙張說明）
        pdesc = (str) 紙張說明
        sdesc = (str) section 說明
        rdesc = (str) 資源說明

    data     = (bytes) 二進位資料
    html     = (str) 採用 html makeup 的內容。但不包括 <html>、<body> 等
               框架元素。

    res_type = (str) 資源附加類別，用在 resource 中。
               採用不同附加方式的 resource，有時可以套用不同的處理。
               注意此處的「類別」與格式無關，非指 png 與 jpg 的差異。

其他變數簡寫法：

    res      = resource
    e        = xml.eTree.ElementTree 的 element 物件

"""

import os.path
import xml.etree.ElementTree as ET
import datetime as DT
import uuid
import collections
import re
import types
import json

import lonote.tar3

# 定義錯誤

class IMError(Exception):
    '''基本錯誤類別，不該直接使用'''
    pass
class IMValueError(IMError):
    '''輸入參數值無法被正確處理時，跳出本錯誤'''
    pass
class IMNotExist(IMError):
    '''操作時發現有重要的東西不存在時，跳出本錯誤'''
    pass
class IMAlreadyExist(IMError):
    '''操作時發現有東西佔住位置擋路時，跳出本錯誤'''
    pass

# ==========================

# 輔助功能

class SafeId:

    invalid_fname = re.compile(r'[/\\:*?"<>|]')

    @classmethod
    def safe_id_assert(cls, xid):
        '''確定 xid 字串為合法檔名（全平台）'''
        if cls.invalid_fname.search(xid):
            info = '"{}" not a valid filename!'.format(xid)
            raise IMValueError(info)

    @classmethod
    def unsafechars_remove(cls, xid):
        '''回傳移除非法字元的 xid 字串'''
        return cls.invalid_fname.sub('', xid)

class TimeStandard:
    '''提供標準時間管理函式組

        所有產生的 datetime 物件都有設定時區
        '''

    format = '%Y-%m-%dT%H:%M:%S.%f%z'

    def local_timezone():
        '''回傳本機的 DT.timezone '''
        error_seconds = 30 # 兩次時間取值可允許的誤差秒數
        delta = DT.datetime.now() - DT.datetime.utcnow()
        delta = delta / (1000000 * 2 * error_seconds) * (1000000 * 2 * error_seconds)
        return DT.timezone(delta)

    def utcnow():
        '''回傳含有時區訊息的現在世界時間'''
        tz = DT.timezone(DT.timedelta(hours = 0))
        return DT.datetime.utcnow().replace(tzinfo = tz)

    @classmethod
    def utcnowtext(cls):
        '''回傳現在世界時間的字串'''
        return cls.totext(cls.utcnow())

    @classmethod
    def now(cls):
        '''回傳含有時區訊息的本地時間'''
        return DT.datetime.now().replace(
                tzinfo = cls.local_timezone())

    @classmethod
    def nowtext(cls):
        '''回傳本地時間的字串'''
        return cls.totext(cls.now())

    @classmethod
    def apply_timezone(cls, time):
        '''回傳一個包含 timezone 的 datatime 物件

            time 為一 datetime 物件

            如果 time 中不包含 timezone 訊息，則套用本地 timezone
            如果已經包含，則不予變更
            '''
        if not time.tzname():
            time = time.replace(tzinfo = cls.local_timezone())
        return time

    @classmethod
    def fromtext(cls, text):
        '''將 text 轉換成 datetime'''
        return DT.datetime.strptime(text, cls.format)

    @classmethod
    def totext(cls, time):
        '''將 datetime 轉換成時間字串'''
        return time.strftime(cls.format)

    def asutc(time):
        '''將 time 轉換成基於 世界標準時 的時間'''
        return time.astimezone(DT.timezone.utc)

    @classmethod
    def aslocal(cls, time):
        '''將 time 轉換成基於 本地時 的時間'''
        return time.astimezone(cls.local_timezone())



# ==========================

# DataBox 資料收集箱

class DataBox(metaclass = collections.abc.ABCMeta):
    '''處理由 xml 管控的資源檔案

        xml 格式為……
            <container attr1="" attr2="" ... >
                <data attr1="" attr2="" ... />
                <data attr1="" attr2="" ... />
            </container>

        data 中會有一項 attr 被指定為 xid
        xid 在一個 box 中需為唯一
        '''

    # base property

    @property
    @collections.abc.abstractmethod
    def _container_name(self):
        '''回傳容器名稱'''
    @property
    @collections.abc.abstractmethod
    def _id_name(self):
        '''回傳 id 名稱'''
    @property
    @collections.abc.abstractmethod
    def _data_name(self):
        '''回傳 data 名稱'''

    # Initialization & Check

    @collections.abc.abstractmethod
    def __init__(self):
        '''初始化，需要透過 super().__init__() 來重用'''
        self.__e_container_load()

    def is_exist(self):
        '''回傳本物件在檔案系統中的對應物是否存在 (True or False)'''
        try:
            self._xml_in_file
            return True
        except IMNotExist:
            return False

    def update(self, frombox):
        '''將自身內容更新為目標 frombox 的內容

            重複或不存在的部份將會從 frombox 讀入並複寫
            self 自身獨有的資料則保持不變。

            frombox 必須要是一個 databox
            '''
        self.meta_update(**frombox.meta)
        for xid in frombox.ids: # 更新 data 項目
            meta = frombox.data_meta(xid)
            meta = { k: meta[k] for k in meta if k != self._id_name }
            data = frombox.data_entity(xid)
            self.data_entity_update(xid,
                    data = data, create = True, **meta)

    # container manage for inner

    @property
    @collections.abc.abstractmethod
    def _path_xml(self):
        '''回傳 xml 路徑'''
    @property
    @collections.abc.abstractmethod
    def _xml_in_file(self):
        '''回傳檔案中的 xml，如果 xml 檔案不存在會 raise IMNotExist'''
    @property
    def _xml_in_memory(self):
        '''回傳記憶體中的 xml'''
        return ET.tostring(self.__e_container, encoding = 'unicode')
    @collections.abc.abstractmethod
    def _xml_delete(self):
        '''刪掉對應的 xml 檔案'''
    @property
    def __new_e_container(self):
        '''回傳新的 e_container'''
        return ET.Element( self._container_name )

    def __e_container_load(self):
        '''從檔案中載入 self.__e_container，如果不存在會自動初始化'''
        try:
            xml = self._xml_in_file
            self.__e_container = ET.fromstring(xml)
        except IMNotExist:
            self.__e_container = self.__new_e_container
    @collections.abc.abstractmethod
    def _e_container_flush(self):
        '''將當前的 self.__e_container 寫入檔案系統中'''

    # container manage for outer

    @property
    def meta(self):
        '''回傳 container 的 attr'''
        return types.MappingProxyType(self.__e_container.attrib)

    def meta_update(self, **kwargs):
        '''更新 container 的 attr

            如果輸入值為 None 且此一 attr 存在，將該 attr 刪除

            如果輸入值不為 str 格式，拋出 IMValueError 錯誤。
            key == self._id_name 時禁止更新，會拋出 IMValueError
            '''
        for k, v in kwargs.items():
            if v is None:
                if k in self.__e_container.attrib:
                    del self.__e_container.attrib[k]
            elif type(v) is not str:
                info = 'Value not allow: must str format!'
                raise IMValueError(info)
            else:
                self.__e_container.set(k, v)
        self._e_container_flush()

    # data meta manage for inner

    def __e_data(self, xid):
        '''回傳 self.__e_container 中，idname == xid 的 _e_data 對象

            不存在則 raise IMNotExist
            '''
        match = './/{data_name}[@{id_name}="{xid}"]'.format(
                data_name = self._data_name,
                id_name = self._id_name,
                xid = xid)
        e_data = self.__e_container.find(match)
        if e_data is None:
            info = '{idname} = "{xid}" are not exist!'.format(
                    idname = self._id_name,
                    xid = xid)
            raise IMNotExist(info)
        else:
            return e_data

    # data meta manage for outer

    @property
    def ids(self):
        '''取得所有 data 的 xid'''
        return { e_data.attrib[self._id_name] for e_data
                in self.__e_container.iter(tag = self._data_name) }

    def data_meta(self, xid):
        '''回傳 dataid = xid 的 data 情報'''
        e_data = self.__e_data(xid)
        return types.MappingProxyType(e_data.attrib)

    def data_meta_update(self, xid, **kwargs):
        '''更新資源情報

            如果輸入值為 None 且此一 attr 存在，將該 attr 刪除

            如果輸入值不為 str 格式，拋出 IMValueError 錯誤。
            key == self._id_name 時禁止更新，會拋出 IMValueError

            注意：只有「已存在」的資源才能被更新情報
                如果資源不存在則拋出 IMNotExist
            '''
        if self._id_name in kwargs: # 企圖修改 id 值時 IMValueError
            info = 'Key "{}" is xid, It cannot be changed!'.format(
                    self._id_name)
            raise IMValueError(info)
        e_data = self.__e_data(xid) # 不存在時，由此拋出 IMNotExist
        for k, v in kwargs.items():
            if v is None: # 刪掉元素
                if k in e_data.attrib:
                    del e_data.attrib[k]
            elif type(v) is not str: # 不屬於 str 則 IMValueError
                info = 'Value not allow: "{}" must str format!'.format(k)
                raise IMValueError(info)
            else:
                e_data.set(k, v)
        self._e_container_flush()

    # data entity manage for inner

    @collections.abc.abstractmethod
    def _path_entity(self, xid):
        '''回傳 idname == xid 的 data 的檔案路徑

            1. 呼叫時，相應的 data_meta 有可能還未產生
                請不要用 e_data 來產生路徑
            2. 請確定不同的 xid 會指向不同的路徑
            '''
    @collections.abc.abstractmethod
    def _data_entity_write(self, xid, enc_data):
        '''寫入 entity 到 self._path_entity(xid) 位置

            此處直接寫入即可，編碼已經透過前端的
            self._data_entity_encode() 完成
            '''
    @collections.abc.abstractmethod
    def _data_entity_load(self, xid):
        '''回傳 xid 指定的 entity

            此處只負責讀出，解碼由
            self._data_entity_decode() 進行
            '''

    def _data_entity_encode(self, xid, data):
        '''回傳編碼後的 data，預設啥也不做直接回傳，請適當複寫該行為

            如果編碼訊息有作為 meta 存在，可以透過 self.data_meta(xid)
            取得當前編碼資料。
            '''
        return data

    def _data_entity_decode(self, xid, enc_data):
        '''回傳解碼後的 data，預設啥也不做直接回傳。請適當複寫該行為

            如果 meta 中有紀錄編碼，建議可由 self.meta(xid) 讀出
            '''
        return enc_data
    @property
    def _data_entity_create_meta_default(self):
        '''回傳一個新建 data 的 meta dict。

            用來在 data_entity_update(create = True) 時使用
            可以預留欄位，設定初始時間、設定預設值等方方面面。
            此處回傳的預設值，遇到重名的狀況，會被創建時輸入的值取代
            '''
        return {}
    @property
    def _data_entity_update_meta_default(self):
        '''回傳用來更新 meta 的 meta dict

            會在 data_entity_update(create = False) 時自動使用，作為預設
            的 meta 值被寫入。

            本身不接受任何參數，但可以用來更新時間戳之類的 meta 屬性
            '''
        return {}

    # data entity manage for outer

    def data_create(self, xid, data, **meta):
        '''創建一個新的 data 項目。

            data_create 會用 self.data_entity_update(create = True) 來新建
            一個 data 項目。

            如果 data 已存在則拋出 IMAlreadyExist
            '''
        try: self.__e_data(xid) # 存在測試
        except IMNotExist: pass
        else:
            info = 'xid = "{}" Cannot create a exist item!'.format(xid)
            raise IMAlreadyExist(info)
        self.data_entity_update(xid, data = data, create = True, **meta)

    def data_entity(self, xid):
        '''回傳指定 data 的 entity 部份

            檔案不存在則 raise IMNotExist
            '''
        path = self._path_entity(xid)
        try:
            enc_data = self._data_entity_load(xid)
            return self._data_entity_decode(xid, enc_data)
        except KeyError:
            info = 'Entity of "{}" Not found!'.format(path)
            raise IMNotExist(info)

    def data_entity_update(self, xid, data, create = False, **meta):
        '''更新 data 的 entity

            create 為真時，data 無論存在或否都不會跳出錯誤，而是有需要就
            會自動新建一個。但不會為它進行任何其他的初始化。

            註：如果需要初始化，請改用 self.data_create()

            當 create == False 時 data 項不存在則跳出 IMNotExist
            '''
        try:
            self.__e_data(xid) # 存在測試
        except IMNotExist as e:
            if not create:
                raise e # 繼續拋出
            # 不存在，初始化 meta
            attrib = collections.ChainMap({self._id_name: xid}, # id 值
                    meta, # 外部輸入 meta
                    self._data_entity_create_meta_default) # 預設 meta
            e_new_data = ET.Element(self._data_name,
                    attrib = dict(attrib))
            self.__e_container.append(e_new_data)
            self._e_container_flush()
        else: # 已存在，更新 meta
            full_meta = collections.ChainMap(meta,
                self._data_entity_update_meta_default)
            self.data_meta_update(xid, **full_meta)
        finally: # 寫入 entity
            enc_data = self._data_entity_encode(xid, data)
            self._data_entity_write(xid, enc_data)

    # delete for inner

    def _data_meta_delete(self, xid):
        '''刪除 data 在 e_container 中的相應內容 '''
        e_data = self.__e_data(xid)
        self.__e_container.remove(e_data)
    @collections.abc.abstractmethod
    def _data_entity_delete(self, xid):
        '''刪除 data entity '''

    # delete for outer

    def data_delete(self, xid):
        '''刪除一筆 data

            如果資料不存在，拋出 IMNotExist
            '''
        self._data_meta_delete(xid)
        self._data_entity_delete(xid)
        self._e_container_flush()

    def data_deleteall(self, box_delete = False):
        '''刪除所有現有的 data

            box_delete == True 表示連同容器也刪除，
            否則將會留下一個空的容器。
            '''
        for xid in self.ids:
            self._data_meta_delete(xid)
            self._data_entity_delete(xid)
        if box_delete == False:
            self._e_container_flush()
        else:
            self._xml_delete()

class TarDataBox(DataBox):
    '''將 DataBox 的資料存在 lonote.tar3.Tar 物件裡'''

    # 繼承來的

    def __init__(self, tar):
        '''將 tar 指派到自身的 self._tar'''
        self._tar = tar
        super().__init__()
    @property
    def _xml_in_file(self):
        '''回傳檔案中的 xml，如果 xml 檔案不存在會 raise IMNotExist'''
        try:
            return self._tar.load(self._path_xml, mode = 'txt')
        except KeyError as e:
            raise IMNotExist(str(e))

    def _e_container_flush(self):
        '''將當前的 self._e_container 寫入檔案系統中'''
        self._tar.update(self._path_xml,
                data = self._xml_in_memory, mode = 'txt')

    def _xml_delete(self):
        '''刪掉對應的 xml 檔案'''
        self._tar.delete(self._path_xml)

    def _data_entity_write(self, xid, enc_data):
        '''寫入 entity 到 self._path_entity(xid) 位置

            此處直接寫入即可，編碼已經透過前端的
            self._data_entity_encode() 完成
            '''
        self._tar.update(self._path_entity(xid),
                data = enc_data, mode = 'bytes')

    def _data_entity_load(self, xid):
        '''回傳 xid 指定的 entity

            此處只負責讀出，解碼由
            self._data_entity_decode() 進行
            '''
        path = self._path_entity(xid)
        return self._tar.load(path, mode = 'bytes')

    def _data_entity_delete(self, xid):
        '''刪除 data entity '''
        self._tar.delete(self._path_entity(xid))

class BookToyBox(TarDataBox):
    '''任意附加資料的儲存箱

        透過特殊 meta 欄位 toy_type 來定義資料格式，達成自動編解碼特性
        新 data 的預設 toy_type 為 binary
        '''

    # 繼承而來加以複寫的內容

    @property
    def _container_name(self):
        '''回傳容器名稱'''
        return 'toybox'
    @property
    def _id_name(self):
        '''回傳 id 名稱'''
        return 'tid'
    @property
    def _data_name(self):
        '''回傳 data 名稱'''
        return 'toy'
    @property
    def _path_xml(self):
        '''回傳 xml 路徑'''
        return 'toybox.xml'

    def _path_entity(self, xid):
        '''回傳 idname == xid 的 data 的檔案路徑

            1. 呼叫時，相應的 data_meta 有可能還未產生
                請不要用 e_data 來產生路徑
            2. 請確定不同的 xid 會指向不同的路徑
            '''
        return 'toybox/' + xid
    @property
    def _data_entity_create_meta_default(self):
        '''回傳一個新建 data 的 meta dict。

            用來在 data_entity_update(create = True) 時使用
            可以預留欄位，設定初始時間、設定預設值等方方面面。
            此處回傳的預設值，遇到重名的狀況，會被創建時輸入的值取代
            '''
        time = TimeStandard.utcnowtext()
        return {'desc': '', 'ctime': time, 'mtime': time,
                'toy_type': 'binary'}
    @property
    def _data_entity_update_meta_default(self):
        '''回傳用來更新 meta 的 meta dict

            會在 data_entity_update(create = False) 時自動使用，作為預設
            的 meta 值被寫入。

            本身不接受任何參數，但可以用來更新時間戳之類的 meta 屬性
            '''
        time = TimeStandard.utcnowtext()
        return {'mtime': time}

    def _data_entity_encode(self, xid, data):
        '''回傳編碼後的 data，預設啥也不做直接回傳，請適當複寫該行為

            如果編碼訊息有作為 meta 存在，可以透過 self.data_meta(xid)
            取得當前編碼資料。
            '''
        t = self.data_meta(xid).get('toy_type')
        if t == 'text': return data.encode('utf8')
        elif t == 'json': return json.dumps(data).encode('utf8')
        else: # 其他狀況不做編碼。需為 bytes 否則會出錯
            return data

    def _data_entity_decode(self, xid, enc_data):
        '''回傳解碼後的 data，預設啥也不做直接回傳。請適當複寫該行為

            如果 meta 中有紀錄編碼，建議可由 self.meta(xid) 讀出
            '''
        t = self.data_meta(xid).get('toy_type')
        if t == 'text': return enc_data.decode('utf8')
        elif t == 'json': return json.loads(enc_data.decode('utf8'))
        else: # 其他狀況不做解碼。
            return enc_data

class ResourceBox(TarDataBox):
    '''供 Paper 使用的附加檔案儲存箱'''

    # 繼承而來加以複寫的內容

    @property
    def _container_name(self):
        '''回傳容器名稱'''
        return 'resources'
    @property
    def _id_name(self):
        '''回傳 id 名稱'''
        return 'rid'
    @property
    def _data_name(self):
        '''回傳 data 名稱'''
        return 'resource'

    def __init__(self, tar, pid):
        self._pid = pid
        super().__init__(tar)
    @property
    def _path_xml(self):
        '''回傳 xml 路徑'''
        return 'pages/' + self._pid + '/resource.xml'

    def _path_entity(self, xid):
        '''回傳 idname == xid 的 data 的檔案路徑

            1. 呼叫時，相應的 data_meta 有可能還未產生
                請不要用 e_data 來產生路徑
            2. 請確定不同的 xid 會指向不同的路徑
            '''
        return 'pages/' + self._pid + '/resources/' + xid
    @property
    def _data_entity_create_meta_default(self):
        '''回傳一個新建 data 的 meta dict。

            用來在 data_entity_update(create = True) 時使用
            可以預留欄位，設定初始時間、設定預設值等方方面面。
            此處回傳的預設值，遇到重名的狀況，會被創建時輸入的值取代
            '''
        time = TimeStandard.utcnowtext()
        return {'desc': '', 'ctime': time, 'mtime': time,
                'res_type': 'att'}
    @property
    def _data_entity_update_meta_default(self):
        '''回傳用來更新 meta 的 meta dict

            會在 data_entity_update(create = False) 時自動使用，作為預設
            的 meta 值被寫入。

            本身不接受任何參數，但可以用來更新時間戳之類的 meta 屬性
            '''
        time = TimeStandard.utcnowtext()
        return {'mtime': time}

class TarTextDataBox(TarDataBox):
    '''資料實體一律以 txt 格式被存取的 TarDataBox'''

    def _data_entity_encode(self, xid, data):
        '''回傳編碼後的 data，預設啥也不做直接回傳，請適當複寫該行為

            如果編碼訊息有作為 meta 存在，可以透過 self.data_meta(xid)
            取得當前編碼資料。
            '''
        return data.encode('utf8')

    def _data_entity_decode(self, xid, enc_data):
        '''回傳解碼後的 data，預設啥也不做直接回傳。請適當複寫該行為

            如果 meta 中有紀錄編碼，建議可由 self.meta(xid) 讀出
            '''
        return enc_data.decode('utf8')

class SectionBox(TarTextDataBox):
    '''供 Paper 使用的頁面部件儲存箱'''

    # 繼承而來加以複寫的內容

    def __init__(self, tar, pid):
        self._pid = pid
        super().__init__(tar)
    @property
    def _container_name(self):
        '''回傳容器名稱'''
        return 'sections'
    @property
    def _id_name(self):
        '''回傳 id 名稱'''
        return 'sid'
    @property
    def _data_name(self):
        '''回傳 data 名稱'''
        return 'section'
    @property
    def _path_xml(self):
        '''回傳 xml 路徑'''
        return 'pages/' + self._pid + '/sections.xml'

    def _path_entity(self, xid):
        '''回傳 idname == xid 的 data 的檔案路徑

            1. 呼叫時，相應的 data_meta 有可能還未產生
                請不要用 e_data 來產生路徑
            2. 請確定不同的 xid 會指向不同的路徑
            '''
        return 'pages/' + self._pid + '/sections/' + xid
    @property
    def _data_entity_create_meta_default(self):
        '''回傳一個新建 data 的 meta dict。

            用來在 data_entity_update(create = True) 時使用
            可以預留欄位，設定初始時間、設定預設值等方方面面。
            此處回傳的預設值，遇到重名的狀況，會被創建時輸入的值取代
            '''
        time = TimeStandard.utcnowtext()
        return {'desc': '', 'ctime': time, 'mtime': time}
    @property
    def _data_entity_update_meta_default(self):
        '''回傳用來更新 meta 的 meta dict

            會在 data_entity_update(create = False) 時自動使用，作為預設
            的 meta 值被寫入。

            本身不接受任何參數，但可以用來更新時間戳之類的 meta 屬性
            '''
        time = TimeStandard.utcnowtext()
        return {'mtime': time}




# ==========================

# Paper 項目

class Paper:
    '''每個筆記頁對應到一個 Paper，此為筆記的最基本單位

        Paper 含有：
            self.resources：
                一個 ResourceBox，用來儲放附加檔案。
            self.sections：
                一個 SectionBox，儲放 html 格式的文件。

        Paper 不該由外部使用者手動建立，而需由 IM 建立並管理
        '''

    default_sid = 'main'

    # Initialization & Check

    def __init__(self, tar, pid):
        self.__resources = ResourceBox(tar, pid)
        self.__sections = SectionBox(tar, pid)
        self.__pid = pid

    def is_exist(self):
        '''回傳這個 pid 頁面存在嗎？

            透過 self.sections 是否存於檔案系統中來判斷
            '''
        return self.__sections.is_exist()

    def create(self, name = 'Unnamed Paper', desc = ''):
        '''在檔案系統中新建此一 Paper

            如果此一 Paper 已存在於檔案系統中，跳出 IMAlreadyExist
            '''
        if self.is_exist():
            info = 'pid: "{}" are already Exist! Cancel!'.format(
                    self.__pid)
            raise IMAlreadyExist(info)
        else:
            self.__sections.meta_update(name = name, desc = desc)
            sid = type(self).default_sid # default section
            self.__sections.data_create(sid, data = '')

    def update(self, from_paper):
        '''用其他 paper 的資料來更新自身

            原有的資料，若未被複寫則保持不變。
            '''
        self.__sections.update(from_paper.sections)
        self.__resources.update(from_paper.resources)

    def delete(self):
        self.__sections.data_deleteall(box_delete = True)
        self.__sections = None
        self.__resources.data_deleteall(box_delete = True)
        self.__resources = None

    # Paper 各部屬性（不含 section 與 resource 內部資料）

    @property
    def pid(self):
        return self.__pid
    @property
    def sections(self):
        return self.__sections
    @property
    def resources(self):
        return self.__resources

    @property
    def name(self):
        '''回傳 paper 標題，映射到 self.sections.meta.get('name')'''
        return self.__sections.meta.get('name')
    @name.setter
    def name(self, name):
        self.__sections.meta_update(name = name)

    @property
    def desc(self):
        '''回傳 paper 說明，映射到 self.sbox.meta.get('desc')'''
        return self.__sections.meta.get('desc')
    @desc.setter
    def desc(self, desc):
        self.__sections.meta_update(desc = desc)



# ==========================

# Book 項目

class Book:
    '''存放有很多 Paper 的活頁簿

        其功能包括：

            1. 作為單獨檔案的最小單位。
                - 也負責開關檔案物件。
            2. 可收藏大量的 Paper，有需要時可將 Paper 取出隨意修改或運用。
            3. 負責排列 Paper 之間的結構。

        關於 Paper 之間的結構紀錄：

            是由以下格式的 xml 檔案來達成的：

                <paper pid="...">
                    <paper pid="..." />
                    <paper pid="...">
                        <paper pid="...">
                            <paper pid="..." />
                        </paper>
                        <paper pid="..." />
                    </paper>
                    <paper pid="..." />
                </paper>

            注意事項包括：
                1. 只有 paper 一種 tag
                2. 每個 paper 都有且只有一個 attribute: pid
                    - 在 Book 介面中，一律透過 pid 來指定特定的 paper
                3. 其排列順序有意義！
                4. 在需要指定位置 (position) 而非元素的場合，用以下格式：
                   (parent_pid, order)
                    - parent_pid  = 一個現存的 pid
                    - order       = parent_pid 下的第幾個，為 0 或正整數
                                  = 0 表示最後一個
                                  = 1 表示第一個，其他正整數依序向後排；
                                    如果超過最大限度，同樣表示最後一個。

            關於最外側的紙張：
                1. 最外側的 paper 只會有一個
                2. 此 paper 被視為書的封面
                    - 其 paper.name（見 Paper 類別）既為 book.name
                    - 其 paper desc == book desc
                3. 全書中，唯有此 paper 不可移動，不可變更位置

        其他紀錄資料有：

            1. 書的實體存放地址 (url)
            2. 用來紀錄其他全書通用資料，具有擴展性的 self.toybox
                - toybox 適合用來紀錄純文字格式的資料，
                  其他格式資料就得自行 base64 了。

        '''

    filename_ext = '.lo3'

    # Check!

    @classmethod
    def url_acceptable_assert(cls, url):
        '''斷言這個 url 可被接受，否則拋出 IMValueError'''
        if not url.endswith(cls.filename_ext):
            info = 'url "{}" not acceptable! must append "{}"!'.format(
                    url, cls.filename_ext)
            raise IMValueError(info)
    @staticmethod
    def url_exist_assert(url):
        '''斷言這個 url 存在，否則拋出 IMNotExist'''
        if not os.path.exists(url):
            info = 'url "{}" not exist!'.format(url)
            raise IMNotExist(info)
    @classmethod
    def url_notexist_assert(cls, url):
        '''斷言這個 url 不存在，否則拋出 IMAlreadyExist'''
        try:
            cls.url_exist_assert(url)
        except IMNotExist:
            pass
        else:
            info = 'url "{}" already exist!'.format(url)
            raise IMAlreadyExist(info)

    # Initialization

    def __variable_init(self):
        '''內部變數初始化'''
        self.__tar = None # lonote.im3.Tar
        self.__path = None
        self.__toybox = None # BookToyBox
        self.__e_tree = None # Element
        self.__papers = None # dict: key = pid, value = paper

    def __init__(self, url, mode = 'load', rootpid = None):
        '''新建書籍物件

            mode 可為 "load" 或 "create"，
                = "load"   代表讀取位在 url 處的現有書籍。
                           書籍不存在會拋出 IMNotExist
                = "create" 代表在 url 處新建書籍。
                           url 位置已經被佔據則拋出 IMAlreadyExist
                其他設值均拋出 IMValueError

            rootpid 只在 create 時有用，這將會指定 root 頁的 pid
                = None 表示自動產生
            '''
        self.__variable_init()
        if mode == 'load':
            self.__load(url)
        elif mode == 'create':
            if rootpid is None:
                rootpid = str(uuid.uuid4())
            self.__create(url, rootpid)
        else:
            info = 'Mode value: "{}" not acceptable!'.format(mode)
            raise IMValueError(info)

    def __load(self, url):
        '''從指定 url 載入資料'''
        def papers_load(self):
            '''以檔案為資料來源，填充 __e_tree 與 __papers '''
            self.__e_tree = ET.fromstring(self.__treexml_in_file)
            self.__papers = { e.get('pid'): Paper(tar = self.__tar, pid = e.get('pid'))
                    for e in self.__e_tree.iter()}

        self.url_acceptable_assert(url)
        self.url_exist_assert(url)
        # Real work
        self.__path = url
        self.__tar = lonote.tar3.Tar(url)
        papers_load(self)
        self.__toybox = BookToyBox(self.__tar)

    def __create(self, url, rootpid):
        '''在指定 url 新建書籍

            rootpid = 首頁的 pid
            '''
        def papers_create(pid):
            '''以預設值填充 __e_tree 與 __papers，此外也寫入檔案中 '''
            paper = self.__paper_new(pid)
            self.__papers = { pid: paper }
            self.__e_tree = self.__e_paper_new(pid)
            self.__e_tree_flush()

        self.url_acceptable_assert(url)
        self.url_notexist_assert(url)
        SafeId.safe_id_assert(rootpid)
        # Real work
        self.__path = url
        self.__tar = lonote.tar3.Tar(url) # 初始化 tar 檔案
        papers_create(rootpid) # 初始化 root paper 和 e_tree
        self.__toybox = BookToyBox(self.__tar)

    def close(self):
        '''關閉檔案物件並重新初始化'''
        self.__tar.close()
        self.__variable_init()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False # 向外傳遞例外

    # property access

    @property
    def path(self):
        '''回傳檔案路徑'''
        return self.__path
    @property
    def bid(self):
        '''回傳 book id ……其實就是檔名（不含副檔名）'''
        return self.filename.rstrip(type(self).filename_ext)
    @property
    def filename(self):
        '''回傳 book 檔名，包含副檔名'''
        return os.path.basename(self.__path)
    @property
    def ids(self):
        '''回傳 book 中所有 pid 列表'''
        return self.__papers.keys()
    @property
    def rootpid(self):
        '''回傳封面 paper 的 pid'''
        return self.__e_tree.get('pid')

    @property
    def name(self):
        '''回傳書本名稱 == 封面 paper 名稱'''
        return self.paper(self.rootpid).name
    @name.setter
    def name(self, value):
        self.paper(self.rootpid).name = value

    @property
    def desc(self):
        '''回傳書本描述 == 封面 paper 描述'''
        return self.paper(self.rootpid).desc
    @name.setter
    def desc(self, value):
        self.paper(self.rootpid).desc = value

    @property
    def toybox(self):
        '''回傳 book 的附加資料箱'''
        return self.__toybox

    # papers tree access

    @property
    def __treexml_path(self):
        '''回傳 treexml 的所在路徑'''
        return 'tree.xml'
    @property
    def __treexml_in_file(self):
        '''回傳檔案中的 treexml，如果檔案不存在會 raise IMNotExist'''
        try:
            return self.__tar.load(self.__treexml_path, mode = 'txt')
        except KeyError as e:
            info = '"{}" not found, this file are invalid!'.format(
                    self.__treexml_path)
            raise IMNotExist(info)
    @property
    def __treexml_in_memory(self):
        '''回傳記憶體中的 treexml'''
        return ET.tostring(self.__e_tree, encoding = 'unicode')

    def __e_tree_flush(self):
        '''將記憶體中的 e_tree 現狀寫入檔案中'''
        self.__tar.update(self.__treexml_path,
                data = self.__treexml_in_memory,
                mode = 'txt')

    # e_paper base operation

    def __e_paper(self, pid):
        '''回傳 self.__e_tree 中，pid 相等的 e_paper 對象

            不存在則 raise IMNotExist
            '''
        if self.__e_tree.get('pid') == pid: # root 測試
            return self.__e_tree
        match = './/paper[@pid="{}"]'.format(pid)
        e_paper = self.__e_tree.find(match)
        if e_paper is None:
            info = 'pid = "{}" are not exist!'.format(pid)
            raise IMNotExist(info)
        else:
            return e_paper

    def __e_paper_parent(self, pid):
        '''取得 parent 的 e_paper'''
        match = './/paper[@pid="{}"]..'.format(pid)
        e_paper_parent = self.__e_tree.find(match)
        if e_paper_parent is None:
            info = '"{}"\'s parent paper are not exist!'.format(pid)
            raise IMNotExist(info)
        else:
            return e_paper_parent

    # e_paper modify operation

    def __position_valid_assert(self, position):
        '''宣稱 position 有效，失敗則拋出 IMError

            有效的條件為是一個二元 tuple 或擁有相同介面的物體
                (parent_pid, order)
                parent_pid 代表的 paper 存在於 self.__e_tree
                order 不可為負數
            '''
        parent_pid, order = position
        if order <= -1:
            info = 'Order not allow Negative!'
            raise IMValueError(info)
        self.__e_paper(parent_pid) # 檢測 parent_pid 存在

    def __paper_detach(self, pid):
        '''將指定的 paper 自主結構樹上脫離，並回傳脫離的 e_paper

            這不會造成 paper 實體檔案內容被刪除，也可能只是在移動過程中

            本操作並不完整，必須要搭配其他操作一併使用
            如徹底刪除或重新插入
            '''
        e_paper = self.__e_paper(pid)
        e_paper_parent = self.__e_paper_parent(pid)
        e_paper_parent.remove(e_paper)
        return e_paper

    def __paper_insert(self, e_paper, position):
        '''將一個 e_paper 項目插入到指定位置'''
        parent_pid, order = position
        e_paper_parent = self.__e_paper(parent_pid)
        if order == 0: # 接在最尾
            e_paper_parent.append(e_paper)
        else: # 大於 0 則依照順序
            # ET 的 insert 是 base 0 的，而我的介面是 base 1，所以要減一
            # 數值超過會自動接在最尾部，所以不用擔心
            e_paper_parent.insert(order - 1, e_paper)

    def __is_lineal(self, ancestor_pid, child_pid):
        '''回傳 ancestor paper 底下是否包含 child paper

            如果 ancestor == child，也算包含（回傳 True）
            '''
        if child_pid == ancestor_pid:
            return True
        e_ancestor = self.__e_paper(ancestor_pid)
        match = './/paper[@pid="{}"]'.format(child_pid)
        e_child = e_ancestor.find(match)
        if e_child is None:
            return False
        else:
            return True

    def paper_move(self, pid, position):
        '''移動 paper 在 tree 中的位置

            pid      = 要移動的 paper
            position = (new_parent_pid, order)

            拋出錯誤：
                pid 不存在、position 格式不對、parent_pid 不存在
                （為 root）、new_parent_pid 不存在、order 為負數。
                以上狀況都會拋出 IMError

            以 order（正整數）為 parent 底下的最終排序順序
                0 == 尾部
                1 == 第一個
                2 == 第二個
                數字小於 0 拋出 IMValueError
                數字大於最大值則直接排到尾巴

            parent 不改變時，以下表作為例子 ( * 為要移動的物件)
                [ 1, 2, 3, *, 5, 6, ... END ]

                to order == 0: 最後
                    [ 1, 2, 3, 5, 6, ... END, * ]
                to order == 1: 最前
                    [ *, 1, 2, 3, 5, 6, ... END ]
                to order == 3:
                    [ 1, 2, *, 3, 5, 6, ... END ]
                to order == 4: 不變
                    [ 1, 2, 3, *, 5, 6, ... END ]
                to order == 5:
                    [ 1, 2, 3, 5, *, 6, ... END, * ]
            '''
        def no_self_reference_assert(pid, new_parent_pid):
            '''斷言不會發生自參照

                如果新的 parent 在自己之下，或等於自己，仍然執行移動，
                則樹狀結構會成為自參照的無限迴圈。
                故需要偵測並避免這種狀況。
                '''
            if self.__is_lineal(
                    ancestor_pid = pid, child_pid = new_parent_pid ):
                info = 'Cannot move to under itself!'
                raise IMValueError(info)
        # Run Test
        new_parent_pid, order = position
        self.__position_valid_assert(position)
        no_self_reference_assert(pid, new_parent_pid)
        # Real work
        e_paper = self.__paper_detach(pid)
        self.__paper_insert(e_paper, position)
        self.__e_tree_flush()

    # e_paper walk operaion

    def pid_children(self, pid):
        '''回傳依照順序的 children pid tuple'''
        return tuple( e.get('pid') for e in list(self.__e_paper(pid)) )

    def pid_parent(self, pid):
        '''回傳 paper 的上級 pid'''
        return self.__e_paper_parent(pid).get('pid')

    def pid_prev(self, pid):
        '''回傳同層級相鄰的上一個 pid, 如果不存在則 raise IMNotExist'''
        e_paper_parent = self.__e_paper_parent(pid)
        e_paper = self.__e_paper(pid)
        children = list(e_paper_parent)
        index = children.index(e_paper)
        newindex = index - 1
        if newindex >= 0:
            return children[newindex].get('pid')
        else:
            info = 'Before paper of "{}" are not exist!'.format(pid)
            raise IMNotExist(info)

    def pid_next(self, pid):
        '''回傳同層級相鄰的下一個 pid, 如果不存在則 raise IMNotExist'''
        e_paper_parent = self.__e_paper_parent(pid)
        e_paper = self.__e_paper(pid)
        children = list(e_paper_parent)
        index = children.index(e_paper)
        newindex = index + 1
        if newindex < len(children):
            return children[newindex].get('pid')
        else:
            info = 'Next paper of "{}" are not exist!'.format(pid)
            raise IMNotExist(info)

    # paper read

    def paper(self, pid):
        '''取出 paper

            pid 不存在會拋出錯誤
            '''
        return self.__papers[pid]

    # paper create

    def __e_paper_new(self, pid):
        '''回傳一個新的 e_paper 元素'''
        return ET.Element('paper', pid = pid)

    def __paper_new(self, pid):
        '''在檔案系統中新建一個 paper，並回傳這個 paper

            如果已經存在會跳出 IMAlreadyExist
            '''
        paper = Paper(tar = self.__tar, pid = pid)
        paper.create()
        return paper

    def __pid_unoccupied_assert(self, pid):
        '''確定 pid 未被佔用，否則 raise IMAlreadyExist'''
        try:
            self.__e_paper(pid)
        except IMNotExist:
            return # 成功
        else:
            info = 'pid: "{}" already exist!'.format(pid)
            raise IMAlreadyExist(info)

    def __pid_new(self):
        '''回傳一個未被使用過的 pid'''
        while True:
            newpid = str(uuid.uuid4())
            try:
                self.__pid_unoccupied_assert(newpid)
                return newpid
            except IMAlreadtExist: pass

    def paper_create(self, position, new_pid = None):
        '''產生全新的 Paper 並將其加入書中，並回傳這個新 paper

            可以指定 new_paper 的 pid，若不指定則會自行產生

            position 表示此一 paper 在書中放置的位置，
            格式為 (parent_pid, order)
            '''
        if new_pid:
            SafeId.safe_id_assert(new_pid)
            self.__pid_unoccupied_assert(new_pid)
        else:
            new_pid = self.__pid_new()
        self.__position_valid_assert(position)
        def get_name(position):
            '''回傳新筆記的名字'''
            # parent_paper = self.paper(position[0])
            # order = position[1]
            # if order == 0:
            #     order = len(self.pid_children(parent_paper.pid)) + 1
            # return parent_paper.name + '-' + str(order) + ' (Unnamed)'
            return 'Unnamed Paper'
        # Real work
        paper = self.__paper_new(new_pid) # create start
        paper.name = get_name(position)
        self.__papers[new_pid] = paper
        e_paper = self.__e_paper_new(new_pid)
        self.__paper_insert(e_paper, position)
        self.__e_tree_flush()
        return paper

    # paper delete

    def paper_delete(self, pid):
        '''刪除本頁面，同時刪除所有次級頁面

            pid 不存在會拋出錯誤
            '''
        e_paper = self.__e_paper(pid)
        e_paper_parent = self.__e_paper_parent(pid)
        pids = [ e.get('pid') for e in e_paper.iter() ]
        for pid in pids:
            paper = self.__papers.pop(pid)
            paper.delete()
        e_paper_parent.remove(e_paper)
        self.__e_tree_flush()

    # paper pid change

    def pid_modify(self, ori_pid, new_pid):
        '''改變 paper 的 pid

            ori_pid 必須存在，new_pid 必須未被佔用，否則都會拋出錯誤。

            未經優化，速度可能比較慢……
            '''
        SafeId.safe_id_assert(new_pid)
        self.__pid_unoccupied_assert(new_pid)
        e_paper = self.__e_paper(ori_pid)
        # Real work
        ori_paper = self.__papers.pop(ori_pid)
        new_paper = Paper(tar = self.__tar, pid = new_pid)
        new_paper.update(ori_paper)
        self.__papers[new_pid] = new_paper
        ori_paper.delete()
        e_paper.set('pid', new_pid)
        self.__e_tree_flush()

    # flush

    def flush(self):
        self.__tar.flush(mode = 'w')

    # export & import & move

    @staticmethod
    def __transfer(src_book, src_pid, dst_book, dst_pid):
        '''將 Paper 內容在書籍之間互相傳輸

            用 src_book 中 src_pid paper 內容，去更新 dst_book 中
            的 dst_pid paper

            兩個 Paper 都需要預先存在
            '''
        src_paper = src_book.paper(src_pid)
        dst_paper = dst_book.paper(dst_pid)
        dst_paper.update(src_paper)
    @classmethod
    def __transfer_tree(cls, src_book, src_pid, dst_book, pidmap):
        '''將整個子樹內容從 src 傳到 dst

            複製過程會包含根結點內容，但 dst_book 中的根 paper
            ( dst_pid == pidmap[src_pid] ) 需要事先存在。

            pidmap 唯一 src_pid 映射到 dst_pid 的字典。
            大略等於：
                {src_pid1: dst_pid1, src_pid2: dst_pid2, ... }

            要複製的所有節點都必須能在 pidmap 中找到 src: dst 對應，否則
            會半途失敗（非原子操作）！
            '''
        dst_pid = pidmap[src_pid]
        cls.__transfer(src_book, src_pid, dst_book, dst_pid)
        for src_child_pid in src_book.pid_children(src_pid):
            dst_child_pid = pidmap[src_child_pid]
            dst_book.paper_create(position = (dst_pid, 0),
                    new_pid = dst_child_pid)
            cls.__transfer_tree(src_book, src_child_pid, dst_book, pidmap)
    @staticmethod
    def __ids_subbook(book, pid):
        '''取得以 pid 為 root 的子書的，含自己在內所有 pid 集合'''
        return { e.get('pid') for e in book.__e_paper(pid).iter() }
    @staticmethod
    def __pidmap_gen(transfer_pids, occupied_pids):
        '''回傳 pidmap，供 __transfer_tree() 使用

            遇到碰撞會自動產生替代的 pid

            transfer_pids = 作為複製來源的 book 中，「需要複製的」pid 
                            集合 (set)。只有打算複製的 pid 需要列入。
                            可以透過 __ids_subbook() 取得
            occupied_pids = 作為複製目標的 book 中，已經存在的 pid 集合
                            這些 pid 有可能擋到 transfer_pid 的路，害
                            transfer_pid 需要更名
            '''
        pidmap = { transfer_pid: transfer_pid
                for transfer_pid in transfer_pids } # default
        conflict_pids = transfer_pids & occupied_pids
        for transfer_pid in conflict_pids:
            while True:
                test_pid = str(uuid.uuid4())
                if test_pid in occupied_pids:
                    continue
                else:
                    pidmap[transfer_pid] = test_pid
                    occupied_pids.add(test_pid)
                    break
        return pidmap

    def book_export(self, url, pid):
        '''將指定的 pid 子結構作為一本新書，匯出到指定的 url 位置去'''
        with Book(url = url, mode = 'create',
                rootpid = pid) as dst_book:
            src_pids = type(self).__ids_subbook(self, pid)
            pidmap = type(self).__pidmap_gen(src_pids, set())
            type(self).__transfer_tree(self, pid, dst_book, pidmap)
            # toybox
            dst_book.toybox.update(self.toybox)

    def book_import(self, src_url, src_pid, dst_position, avoid_conflict = False):
        '''將 url 處的 book 匯入到自身的 position 位置，作為新增的子結構

            avoid_conflict == True 則遇到 pid 衝突會自行解決（自動變更
                              衝突的 pid）
                              否則偵測到 id 衝突會自動報錯
            '''
        with Book(src_url, mode = 'load') as src_book:
            # 先解決 pid 衝突問題
            occupied_pids = self.ids
            transfer_pids = type(src_book).__ids_subbook(src_book, src_pid)
            conflict_pids = occupied_pids & transfer_pids # 衝突項目
            if conflict_pids: # 存在衝突
                if not avoid_conflict:
                    info = 'pid has some conflict: {}'.format(
                            conflict_pids)
                    raise IMAlreadyExist(info)
            # Real work
            pidmap = self.__pidmap_gen(transfer_pids, occupied_pids)
            self.paper_create( # root create first
                    position = dst_position, new_pid = pidmap[src_pid])
            type(self).__transfer_tree(src_book, src_pid, self, pidmap)
