#coding=utf-8

###################################################################
# Copyright 2013 Civa lin 林雪凡
# 
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#      http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
###################################################################

"""
LoNote translate 翻譯模組

定義翻譯功能 API。
"""

import gettext
import functools
import locale
import os.path
import sys

def workaround_bottle_import_error_in_win32():
    '''以下程式碼被用來在 Win32GUI 環境下繞過 bottle import 錯誤'''
    try:
        sys.stdout.write("")
        sys.stdout.flush()
    # TESTED: Should use AttributeError replace IOError in python 3.3
    except AttributeError:
        class dummyStream:
            ''' dummyStream behaves like a stream but does nothing. '''
            def __init__(self): pass
            def write(self,data): pass
            def read(self,data): pass
            def flush(self): pass
            def close(self): pass
        # and now redirect all default streams to this dummyStream:
        sys.stdout = dummyStream()
        sys.stderr = dummyStream()
        sys.stdin = dummyStream()
        sys.__stdout__ = dummyStream()
        sys.__stderr__ = dummyStream()
        sys.__stdin__ = dummyStream()

workaround_bottle_import_error_in_win32()
import bottle


_LOCALE_DIR = None # None 表示使用系統預設位置，否則必須從外部進行設定

def set_locale_dir(locale_dir):
    '''如果指定的資料夾不存在，就自動設為系統預設位置'''
    if os.path.isdir(locale_dir):
        global _LOCALE_DIR
        _LOCALE_DIR = locale_dir

def _getlocale_codes(request):
    '''從 bottle.request 中取得一個列表，其中記錄了期望使用的語言順序'''
    def normalize(code):
        '''把取得的locale名稱字串大小寫固定為標準的 zh_TW 這種格式'''
        frags = code.split('-')
        if len(frags) == 2:
            ans = frags[0].lower() + '_' + frags[1].upper()
        else:
            ans = frags[0].lower()
        return ans
    codes = request.environ.get("HTTP_ACCEPT_LANGUAGE", "").split(";")[0].split(',')
    return (i for i in map(normalize, codes))

@functools.lru_cache()
def get_trans(codes):
    '''依照順序套用可用的 codes'''
    try:
        trans = gettext.translation('lonote', _LOCALE_DIR, languages = codes)
        return trans
    except (IOError, OSError):
        return gettext.NullTranslations()

def get_server_gettext_func():
    '''使用 server 預設語系來取得 gettext function'''
    try:
        lang, enc = locale.getdefaultlocale()
        trans = gettext.translation('lonote', _LOCALE_DIR, languages = [lang])
        return trans.gettext
    except (IOError, OSError):
        return gettext.gettext

def install():
    '''將當前語系安裝起來'''
    codes = _getlocale_codes(bottle.request)
    trans = get_trans(codes)
    trans.install()

def renew(func):
    '''函式裝飾器，裝飾後的函式可以自動重新安裝 gettext 到 _'''
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        install()
        return func(*args, **kwargs)
    return wrapper
